#include <stdio.h>
#include <math.h>
//#include <values.h>
#include <gdk/gdk.h>
#include "gtkgauge.h"
#include "gtkhst-compat.h"
#include "axis.h"

PangoLayout *_hst_get_pango_layout(GtkWidget *w, cairo_t *cr, const char *font_family);
void _hst_draw_string (GtkWidget *w, cairo_t *cairo,
                      gint x,
                      gint y,
                      const char *s,
                      const char *font_family);
void _hst_get_layout_extent(GtkWidget *w, cairo_t *, const char *text, PangoRectangle *fsize, const char *font);

struct _GtkGaugePrivate {
        int ox, oy;
        int rx, ry;
        int cx, cy;
        double cdiag;
        gdouble pannel;  /* fraction of the area colored */
        gdouble  angmn;  /* minimum angle at which the scale is drawn */
        gdouble  step;
        gdouble  dot;
        gdouble  sx, ex;
        cairo_t *cairo;
        cairo_surface_t *surface;
        GdkFont *ticks;
        GdkFont *label;
        GdkFont *upper;
        char *font_family;
        GdkColor color;
        GMutex   *mutex;
};


static GtkDrawingAreaClass *parent_class;

static void   gauge_class_init (GtkGaugeClass * klass);
static void   gauge_init (GtkGauge * gg);
static void   gauge_size_allocate (GtkWidget * widget, GtkAllocation * allocation);
static void   gauge_size_request (GtkWidget *, GtkRequisition *);
static void   gauge_update_scale (GtkGauge * gg);
static gint   gauge_expose (GtkWidget * widget, GdkEventExpose * event);
static void   gauge_paint (GtkGauge *G, cairo_t *cairo);
static void   gauge_finalize (GObject *);
static double gauge_get_phi (GtkGauge * gg, double);
static void   gauge_set_property        (GObject	 *object,
                                         guint           property_id,
                                         const GValue   *value,
                                         GParamSpec     *pspec);
static void   gauge_get_property        (GObject	 *object,
                                         guint           property_id,
                                         GValue         *value,
                                         GParamSpec     *pspec);

enum {
    PROP_0,
    PROP_XMIN,
    PROP_XMAX,
    PROP_TITLE,
    PROP_COLOR,
    PROP_FONT,
};


GType gtk_gauge_get_type (void)
{
    static GType gauge_type = 0;

    if (!gauge_type) {
        static const GTypeInfo gauge_info = {
                                             sizeof(GtkGaugeClass),
                                             NULL,               /* base_init */
                                             NULL,               /* base_finalize */
                                             (GClassInitFunc) gauge_class_init,
                                             NULL,               /* class_finalize */
                                             NULL,               /* class_data */
                                             sizeof(GtkGauge),
                                             0,          /* n_preallocs */
                                             (GInstanceInitFunc) gauge_init,
        };

        gauge_type = g_type_register_static (GTK_TYPE_DRAWING_AREA,
                                             "GtkGauge",
                                             &gauge_info,
                                             0);
    }

    return gauge_type;
}

static void gauge_class_init (GtkGaugeClass * class)
{
    GObjectClass   *gobject_class;
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;

    parent_class = g_type_class_peek_parent (class);

    gobject_class = G_OBJECT_CLASS (class);
    object_class = GTK_OBJECT_CLASS (class);
    widget_class = GTK_WIDGET_CLASS (class);

    gobject_class->finalize     = gauge_finalize;
    gobject_class->set_property = gauge_set_property;
    gobject_class->get_property = gauge_get_property;

    widget_class->size_allocate = gauge_size_allocate;
    widget_class->expose_event = gauge_expose;
    widget_class->size_request = gauge_size_request;

    g_object_class_install_property (
            gobject_class,
            PROP_XMIN,
            g_param_spec_double("xmin","xmin","Minimum value of X",
                                -DBL_MAX, DBL_MAX,0.,
                                G_PARAM_READABLE | G_PARAM_WRITABLE)
    );
    g_object_class_install_property (
            gobject_class,
            PROP_XMAX,
            g_param_spec_double("xmax","xmax","Maximum value of X",
                                -DBL_MAX, DBL_MAX,1.,
                                G_PARAM_READABLE | G_PARAM_WRITABLE)
    );

    g_object_class_install_property (
            gobject_class,
            PROP_TITLE,
            g_param_spec_string ("title", "Graph Title",
                                 "Graph Title",
                                 "Title",
                                 G_PARAM_READABLE | G_PARAM_WRITABLE)
            );
    g_object_class_install_property (
            gobject_class,
            PROP_COLOR,
            g_param_spec_string ("color", "Graph Color",
                                 "Graph Color",
                                 "Color",
                                 G_PARAM_READABLE | G_PARAM_WRITABLE)
            );
    g_object_class_install_property (
                gobject_class,
                PROP_FONT,
                g_param_spec_string ("font", "Font",
                                     "Font",
                                     "Font",
                                     G_PARAM_READABLE | G_PARAM_WRITABLE)
                );
}

static void gauge_init (GtkGauge * gg)
{
    // TODO: Implement the private part
    gg->priv = g_new0(GtkGaugePrivate, 1);
    gg->priv->mutex = g_mutex_new();
    gg->priv->font_family = g_strdup("sans");
    gg->priv->cairo = 0;
    gg->priv->surface = 0;
    gg->val = 0.;
    gg->xmin = 0.;
    gg->xmax = 1.;
    gg->priv->pannel = 0.2;
    gg->priv->dot = 0.05;
    gg->title = 0;
    gg->squared = FALSE;
    gtk_gauge_set_color(gg,"red");
}

static void
gauge_set_property (GObject      *object,
                    guint         property_id,
                    const GValue *value,
                    GParamSpec   *pspec)
{
    GtkGauge *ff;
    g_return_if_fail( GTK_IS_GAUGE(object) );

    ff = GTK_GAUGE(object);
    switch (property_id) {
        case PROP_XMIN:
            ff->xmin = g_value_get_double(value);
            break;
        case PROP_XMAX:
            ff->xmax = g_value_get_double(value);
            break;
        case PROP_TITLE:
            gtk_gauge_set_title(ff, g_value_get_string(value ) );
            break;
        case PROP_COLOR:
            gtk_gauge_set_color(ff, g_value_get_string(value));
            break;
        case PROP_FONT:
            if (ff->priv->font_family)
                g_free(ff->priv->font_family);
            ff->priv->font_family = g_strdup(g_value_get_string(value));
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
gauge_get_property (GObject     *object,
                    guint        property_id,
                    GValue      *value,
                    GParamSpec  *pspec)
{
    GtkGauge *ff;
    g_return_if_fail( GTK_IS_GAUGE(object) );

    ff = GTK_GAUGE(object);

    switch (property_id) {
        case PROP_XMIN:
            g_value_set_double(value,ff->xmin);
            break;
        case PROP_XMAX:
            g_value_set_double(value,ff->xmax);
            break;
        case PROP_TITLE:
            g_value_set_string(value, ff->title);
            break;
        case PROP_COLOR:
            g_value_set_string(value, ff->color);
            break;
        case PROP_FONT:
            g_value_set_string(value, ff->priv->font_family);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

GtkWidget *gtk_gauge_new (float xmin, float xmax)
{
    GtkGauge *gg = g_object_new ( GTK_TYPE_GAUGE,
                                  "xmin",xmin,
                                  "xmax", xmax,
                                  NULL);

    return GTK_WIDGET (gg);
}

void gauge_finalize (GObject * obj)
{
    GtkGauge *gg;
    g_return_if_fail (obj != NULL);
    g_return_if_fail (GTK_IS_GAUGE (obj));

    gg = GTK_GAUGE (obj);

    if (gg->priv->surface)
    {
        cairo_surface_destroy(gg->priv->surface);
        gg->priv->surface = 0;
    }

    if (gg->title) {
        g_free (gg->title);
        gg->title=0;
    }
    if (gg->color)
    {
        g_free(gg->color);
        gg->color=0;
    }
    g_mutex_free(gg->priv->mutex);
    g_free(gg->priv);
    gg->priv=0;
    G_OBJECT_CLASS (parent_class)->finalize (obj);
}

static inline double gauge_get_phi (GtkGauge * gg, double x)
{
    double v;
    if ( x>=gg->xmin && x<=gg->xmax )
        v = x;
    else if ( x<gg->xmin) {
        v=gg->xmin;
    }
    else
        v=gg->xmax;

    return gg->priv->angmn - gg->priv->step * (v - gg->xmin);
}



void gtk_gauge_set_value (GtkGauge * gg, float v)
{
    gg->val = v;
    gtk_widget_queue_draw (GTK_WIDGET (gg));
}

float gtk_gauge_get_value (GtkGauge * gg)
{
    return gg->val;
}

void gtk_gauge_set_title (GtkGauge * gg, const char *tit)
{
    if (gg->title)
        g_free (gg->title);
    gg->title = (tit != 0 ? g_strdup (tit) : 0);
    if (gtk_widget_get_realized (GTK_WIDGET (gg)))
        gtk_widget_queue_resize (GTK_WIDGET (gg));
}

char *gtk_gauge_get_title (GtkGauge * gg)
{
    return gg->title;
}

void gtk_gauge_set_range (GtkGauge * gg, float xmin, float xmax)
{
    gg->xmin = xmin;
    gg->xmax = xmax;
    gauge_update_scale (gg);
    gtk_widget_queue_draw (GTK_WIDGET (gg));
}

void gtk_gauge_set_color (GtkGauge * gg, const char *color_name)
{
    GdkColor col;
    GtkWidget *widget;

    g_return_if_fail( gg!=NULL );
    g_return_if_fail( GTK_IS_GAUGE(gg) );

    if (! gdk_color_parse(color_name,&col) ) {
        g_printerr("Cannot get color %s\n",color_name);
        return;
    }
    if (gg->color)
        g_free(gg->color);

    gg->color = g_strdup(color_name);
    gg->priv->color = col;

    widget = GTK_WIDGET(gg);
    if ( gtk_widget_get_realized ( widget ) )
        gdk_color_alloc (gtk_widget_get_colormap (widget), &gg->priv->color);
}




static void gauge_size_request (GtkWidget * w, GtkRequisition * requisition)
{
    PangoLayout *layout;
    PangoRectangle logical_rect;

    if (w == NULL)
        return;
    if (!GTK_IS_GAUGE (w))
        return;
    if (requisition == NULL)
        return;

    layout = gtk_widget_create_pango_layout (w, "X");
    pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
    requisition->width  = (int)(17.5*logical_rect.width);
    requisition->height = requisition->width;
    g_object_unref (layout);
    /* Chain up to default */
    GTK_WIDGET_CLASS (parent_class)->size_request (w, requisition);
}


static void gauge_size_allocate (GtkWidget * widget, GtkAllocation * allocation)
{
    GtkGauge *gg;
    cairo_t *cairo;
    PangoRectangle  fsize;
    PangoLayout *layout;
    int ch, cw;
    int cx, cy, ox, oy, rx,ry;
    int  x_offset,y_offset;
    double cdiag;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_GAUGE (widget));
    g_return_if_fail (allocation != NULL);

    /* Chain up to default that simply reads current requisition */
    GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation);

    gg = GTK_GAUGE(widget);

    g_mutex_lock(gg->priv->mutex);
    if (gg->priv->surface)
        cairo_surface_destroy(gg->priv->surface);

    gg->priv->surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
                                                   allocation->width,
                                                   allocation->height);

    /*
     *  Get the center and the radius
     */
    /* Get the size of a character */
    layout = gtk_widget_create_pango_layout (widget, "X");
    pango_layout_get_pixel_extents (layout, NULL, &fsize);
    ch    = fsize.height;
    cw    = fsize.width;
    cdiag = sqrt (ch * ch + cw * cw);

    /* Get dimensions and center of gauge */
    cx = allocation->width;
    cy = allocation->height;
    x_offset = y_offset = 0;
    if ( gg->squared ) {
        if (cx>cy) {
            x_offset = (cx-cy)/2;
            cx = cy;
        }
        else {
            y_offset = (cy-cx)/2;
            cy = cx;
        }
    }
    ox = x_offset + cx / 2;
    oy = y_offset + cy / 2;
    ry = rx = MIN(cx/2,cy/2);

    gg->priv->ox    = ox;
    gg->priv->oy    = oy;
    gg->priv->rx    = rx;
    gg->priv->ry    = ry;
    gg->priv->cx    = cx;
    gg->priv->cy    = cy;
    gg->priv->cdiag = cdiag;


    g_mutex_unlock(gg->priv->mutex);
    if ( gtk_widget_get_realized(widget))
        gauge_expose(widget, 0);
}


static void gauge_paint (GtkGauge *gg, cairo_t *cairo)
{
    GtkStyle *style;
    static char line[50];
    double xp;
    int xx, yy;
    GdkColor *col;
    PangoLayout *layout;
    PangoRectangle logical_rect;

    g_return_if_fail (gg != NULL);

    g_mutex_lock(gg->priv->mutex);
    if (!cairo)
        cairo = gg->priv->cairo = cairo_create(gg->priv->surface);
    else
        gg->priv->cairo = cairo;

    style = gtk_widget_get_style(GTK_WIDGET(gg));
    /*
     * Clear the canvas
     */
    col = style->bg+GTK_STATE_NORMAL;
    cairo_set_source_rgb(cairo, col->red/65535.0, col->green/65535.0, col->blue/65535.0);
    cairo_paint(cairo);

    col = style->fg+GTK_STATE_NORMAL;
    cairo_set_source_rgb(cairo, col->red/65535.0, col->green/65535.0, col->blue/65535.0);

    cairo_set_line_cap(cairo, CAIRO_LINE_CAP_ROUND);
    cairo_set_line_join(cairo, CAIRO_LINE_JOIN_ROUND);

    gauge_update_scale(gg);
    xp = gauge_get_phi (gg, gg->val);
    xx =  (int) (gg->priv->rx * cos (xp) + .5);
    yy = -(int) (gg->priv->ry * sin (xp) + .5);
    cairo_move_to(cairo, gg->priv->ox, gg->priv->oy);
    cairo_line_to(cairo, gg->priv->ox+xx, gg->priv->oy+yy);
    cairo_stroke(cairo);

    sprintf (line, "%.1f", gg->val);
    layout = _hst_get_pango_layout(GTK_WIDGET(gg), cairo, "sans");
    pango_layout_set_markup(layout,line,-1);
    pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
    cairo_move_to(cairo, gg->priv->ox-logical_rect.width/2,gg->priv->oy+logical_rect.height/2.);
    pango_cairo_show_layout(cairo, layout);
    g_object_unref(layout);

    cairo_destroy(cairo);
    gg->priv->cairo = 0;
    g_mutex_unlock(gg->priv->mutex);
}



static gint gauge_expose (GtkWidget * widget, GdkEventExpose * event)
{
    g_return_val_if_fail (widget != NULL, FALSE);
    g_return_val_if_fail (GTK_IS_GAUGE (widget), FALSE);

    cairo_t *cr = gdk_cairo_create(gtk_widget_get_window(widget));
    if (event)
    {
        cairo_rectangle(cr,
                        event->area.x, event->area.y,
                        event->area.width, event->area.height);
        cairo_clip(cr);
    }
    gauge_paint (GTK_GAUGE(widget), cr);
    return FALSE;
}


static void gauge_update_scale (GtkGauge * gg)
{
    GtkWidget *w;
    cairo_t *cairo = gg->priv->cairo;
    int Ndiv[2] = { 10, 5 }, nBins[2];
    float Wmn[2]={0.,0.}, Wmx[2]={0.,0.}, Bw[2] = { 1., 1. };
    double angmn, angRange, pc, pc2, pcr, pcr2;
    double x, y, xp;
    int xo, yo, xx, yy,ptit;
    double sx,ex;
    double R;

    double eps = 0.00001;
    static char line[50];
    GdkColor *col;

    PangoLayout    *layout;
    PangoRectangle  fsize;
    GtkAllocation   allocation;
    GtkStyle *style;

    g_return_if_fail (gg != NULL);
    g_return_if_fail (GTK_IS_GAUGE (gg));

    w = GTK_WIDGET (gg);
    gtk_widget_get_allocation(w, &allocation);
    style = gtk_widget_get_style(GTK_WIDGET(gg));


    /* Get the order of magnitude of the range */
    get_order_of_magnitude (gg->xmax, 3., &sx, &ex);
    sx = (ex!=0. ? pow (10., ex) : 1.);


    /* Get the binning */
    getbins (gg->xmin, gg->xmax, Ndiv[0], Wmn, Wmx, nBins, Bw);
    if ((gg->xmin - Wmn[0]) > eps) {
        Wmn[0] += Bw[0];
        nBins[0]--;
    }
    if ((Wmx[1] - gg->xmax) > eps) {
        Wmx[0] -= Bw[0];
        nBins[0]--;
    }
    getbins (0, Bw[0], Ndiv[1], Wmn + 1, Wmx + 1, nBins + 1, Bw + 1);

    /*
      Get minumun angle and angle range for the scale.
      We only draw the panel from the diagonal of the center to the
      lower left corner until the diagonal of the center to the lower
      right corner.
     */
    angmn    = atan2 ((double) gg->priv->rx, (double) gg->priv->ry);
    angRange = 2. * (M_PI - fabs (angmn));
    angmn    = atan2 (-(double) gg->priv->ry, -(double) gg->priv->rx);
    if (angmn < 0)
        angmn += 2 * M_PI;

    pc        = gg->priv->pannel;
    pc2       = pc / 2;
    pcr       = (1.-pc);
    pcr2      = (1.-pc2);
    gg->priv->angmn = angmn;
    gg->priv->step  = angRange / (gg->xmax - gg->xmin);

    /*
       Build the region that will limit the angular extent of the
       scale.
       (0,0)               (cx,0)
           +--------------+
		   | 	       	  |
		   | 	       	  |
		   | 	(ox,oy)	  |
		   | 	  x    	  |
		   |    .   .	  |
		   |  .	      .	  |
		   |.  	       	. |
		   +--------------+
	       (0,cy)              (cx,cy)

     */
    cairo_save(cairo);
    cairo_move_to(cairo, gg->priv->ox-gg->priv->rx, gg->priv->oy+gg->priv->ry);
    cairo_line_to(cairo, gg->priv->ox-gg->priv->rx, gg->priv->oy-gg->priv->ry);
    cairo_line_to(cairo, gg->priv->ox+gg->priv->rx, gg->priv->oy-gg->priv->ry);
    cairo_line_to(cairo, gg->priv->ox+gg->priv->rx, gg->priv->oy+gg->priv->ry);
    cairo_line_to(cairo, gg->priv->ox, gg->priv->oy);
    cairo_clip(cairo);

    /*
     * Draw and fill the circle that will make the panel
     */
    col = &gg->priv->color;
    R = gg->priv->rx;
    cairo_set_source_rgb(cairo, col->red/65535.0, col->green/65535.0, col->blue/65535.0);
    cairo_arc(cairo, gg->priv->ox, gg->priv->oy, R, 0, 2*M_PI );
    cairo_fill(cairo);

    /*
     * draw main scale
     */
    col = style->fg+GTK_STATE_NORMAL;
    cairo_set_source_rgb(cairo, col->red/65535.0, col->green/65535.0, col->blue/65535.0);

    R = gg->priv->rx;
    for (x = Wmn[0]; x <= Wmx[0]; x += Bw[0]) {
        xp = gauge_get_phi (gg, x);
        xx =  (int) (R * cos (xp) + .5);
        yy =  -(int) (R * sin (xp) + .5);
        cairo_move_to(cairo, gg->priv->ox, gg->priv->oy);
        cairo_line_to(cairo, gg->priv->ox + xx, gg->priv->oy + yy);
        cairo_stroke(cairo);
    }


    R = pcr2*gg->priv->rx;
    /* draw the intermadiate ticks */
    for (x = Wmn[0]; x < Wmx[0]; x += Bw[0]) {
        for (y = Wmn[1]; y < Wmx[1]; y += Bw[1]) {
            if (x == y)
                continue;
            xp = gauge_get_phi (gg, x + y);
            xx = (int) (R * cos (xp) + .5);
            yy = -(int) (R * sin (xp) + .5);
            cairo_move_to(cairo, gg->priv->ox, gg->priv->oy);
            cairo_line_to(cairo, gg->priv->ox + xx, gg->priv->oy + yy);
            cairo_stroke(cairo);
        }
    }

    /*
     * Back to normal clipping
     */
    cairo_restore(cairo);
    col = style->bg+GTK_STATE_NORMAL;
    cairo_set_source_rgb(cairo, col->red/65535.0, col->green/65535.0, col->blue/65535.0);

    R=pcr*gg->priv->rx;
    cairo_arc(cairo, gg->priv->ox, gg->priv->oy, R, 0, 2*M_PI);
    cairo_fill(cairo);
    col = style->fg+GTK_STATE_NORMAL;
    cairo_set_source_rgb(cairo, col->red/65535.0, col->green/65535.0, col->blue/65535.0);

    R = gg->priv->dot*gg->priv->rx;
    cairo_arc(cairo, gg->priv->ox, gg->priv->oy, R, 0, 2*M_PI);
    cairo_fill(cairo);


    /* draw tick labels */
    R = pcr * gg->priv->rx - gg->priv->cdiag;
    for (x = Wmn[0]; x <= Wmx[0]; x += Bw[0]) {
        xp = gauge_get_phi (gg, x);
        sprintf (line, "%.1f", x / sx);
        layout = _hst_get_pango_layout(GTK_WIDGET(gg), cairo, gg->priv->font_family);
        pango_layout_set_markup(layout,line,-1);
        pango_layout_get_pixel_extents (layout, NULL, &fsize);
        xo = (int) ( R * cos (xp) - fsize.width/2. );
        yo = (int) ( R * sin (xp) + fsize.height/2. );
        if ( fabs(xp-M_PI_2)< 0.018 )
            yo += (int)(0.5*fsize.height);
        cairo_move_to(cairo, gg->priv->ox+xo, gg->priv->oy-yo);
        pango_cairo_show_layout(cairo, layout);
        g_object_unref(layout);
    }

    /*
       Draw the title
     */
    ptit = gg->priv->cy;
    if (gg->title) {
        sprintf (line, "<b>%s</b>",gg->title);
        layout = _hst_get_pango_layout(GTK_WIDGET(gg), cairo, gg->priv->font_family);
        pango_layout_set_markup(layout,line,-1);
        pango_layout_get_pixel_extents (layout, NULL, &fsize);
        ptit -= 2+fsize.height;
        cairo_move_to(cairo, gg->priv->ox - fsize.width/2, ptit);
        pango_cairo_show_layout(cairo, layout);
        g_object_unref(layout);
    }
    /*
       Draw the order of magnitude
     */
    if (sx != 1.) {
        sprintf (line, "x10<sup>%d</sup>",(int)ex);
        layout = _hst_get_pango_layout(GTK_WIDGET(gg), cairo, gg->priv->font_family);
        pango_layout_set_markup(layout,line,-1);
        pango_layout_get_pixel_extents (layout, NULL, &fsize);
        ptit -= 2+fsize.height;
        cairo_move_to(cairo, gg->priv->ox - fsize.width / 2, ptit);
        pango_cairo_show_layout(cairo, layout);
        g_object_unref(layout);
    }


}

void
gtk_gauge_set_squared(GtkGauge *gg, gboolean squared)
{
    g_return_if_fail (gg != NULL);
    g_return_if_fail (GTK_IS_GAUGE (gg));

    gg->squared = squared;
    gtk_widget_queue_draw(GTK_WIDGET(gg));
}
