/*
 * tstCairo.c
 *
 *  Created on: Jan 21, 2015
 *      Author: lacasta
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <gtk/gtk.h>
#include <gtk-hst/histogram.h>
#include <gtk-hst/gtkhistogram.h>


static int __init__=0;
static void do_init() {
    __init__=1;
    srand( time(0) );
}

double randuniform() {
    double x;
    if (!__init__)
        do_init();
    x = ((double)rand())/((double)RAND_MAX);
    return x;
}

double randgauss(double mean,double stdev) {
    static int haveCachedNormal = 0;
    static double  cachedNormal = 0.;
    static double  pMean=0.,pStdDev=0.;
    if (!__init__)
        do_init();

    if ( mean!=pMean || pStdDev!=stdev ) {
        haveCachedNormal = 0;
        pMean = mean;
        pStdDev=stdev;
    }

    if (haveCachedNormal == 1) {
        haveCachedNormal = 0;
        return(cachedNormal * pStdDev + pMean );
    } else {
        for(;;) {
            double u1 = randuniform();
            double u2 = randuniform();
            double v1 = 2 * u1 - 1;
            double v2 = 2 * u2 - 1;
            double w = (v1 * v1) + (v2 * v2);

            /*
               We actually generate two IID normal distribution variables.
               We cache the one & return the other.
             */
            if (w <= 1) {
                double y = sqrt( (-2 * log(w)) / w);
                double x1 = v1 * y;
                double x2 = v2 * y;

                haveCachedNormal = 1;
                cachedNormal = x2;
                return(x1 * pStdDev + pMean);
            }
        }
    }
    return 0;
}

gboolean do_print_hst(GtkHst1D *hst)
{
    FILE *of;
    GString *str = 0;
    g_object_get(G_OBJECT(hst), "png", &str,NULL);

    of = fopen("/tmp/kkdvk.png","w");
    if (!of)
    {
        printf("I could not open the output file for writing\n");
    }
    else
    {
        fwrite(str->str, str->len, 1, of);
        fclose(of);
    }
    g_string_free(str, TRUE);
    return TRUE;
}

gboolean
idle_func (HstHistogram *hst)
{
    double val = randgauss(0.0, 8);
    hst_histogram_fill(hst, val, 1);
    return TRUE;
}


GtkWidget *create_mainWindow()
{
    GtkWidget *mainWindow;
    GtkWidget *notebook;
    GtkWidget *vbox;
    GtkWidget *label;
    GtkWidget *hst;
    HstHistogram *hf;
    mainWindow = gtk_dialog_new_with_buttons(
                             "DebugCairo", NULL,
                             GTK_DIALOG_MODAL| GTK_DIALOG_DESTROY_WITH_PARENT,
                             GTK_STOCK_OK,
                             GTK_RESPONSE_ACCEPT,
                             GTK_STOCK_EXECUTE,
                             GTK_RESPONSE_APPLY,
                             NULL
        );

    notebook = gtk_notebook_new();
    vbox = gtk_vbox_new(TRUE, 0);


    gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new("A label"), FALSE, FALSE, 0);
    label = gtk_label_new("tab 1");
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, label);


    hf = hst_histogram_new(50, -50, 50, 0);
    g_timeout_add(1000, (GSourceFunc)idle_func, hf);

    hst = gtk_hst1d_new(HST_FUNCTION(hf));
    gtk_hst1d_monitor_start(GTK_HST1D(hst),500);


    vbox = gtk_vbox_new(TRUE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), hst, FALSE, FALSE, 0);
    label = gtk_label_new("tab 2");
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, label);


    gtk_container_add (GTK_CONTAINER (  gtk_dialog_get_content_area( GTK_DIALOG(mainWindow) )), notebook);

    gtk_widget_show_all(mainWindow);
    g_object_set_data(G_OBJECT(mainWindow), "hst", hst);

    return mainWindow;
}

int main(int argc, char **argv)
{
    GtkWidget *mainWindow;
    gtk_init (&argc, &argv);

    mainWindow = create_mainWindow ();

    int rc;
    do
    {
        rc = gtk_dialog_run(GTK_DIALOG(mainWindow));
        if ( rc == GTK_RESPONSE_APPLY )
        {
            GtkHst1D *hst = GTK_HST1D( g_object_get_data(G_OBJECT(mainWindow), "hst") );
            do_print_hst(hst);
        }
    } while ( rc != GTK_RESPONSE_ACCEPT );
    return 0;
}




