#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <gtk/gtk.h>

#include <gtk-hst/gtkhistogram2d.h>
#include <gtk-hst/tracer.h>
#include <gtk-hst/histogram.h>
#include <gtk-hst/gtkhistogram.h>
#ifdef __APPLE__
# include <mach/task.h>
# include <mach/mach_init.h>
#endif
#include "random.h"

GtkWidget *create_mainWindow (void);

GPrintFunc std_logger;
void my_logger(const gchar *string) {
    printf("[mylogger]%s\n",string);
}




static char sbuf[1024];
int read_file()
{
#ifdef G_PLATFORM_WIN32
    return -1;
#else
    sprintf(sbuf,"/proc/%d/statm", getpid());
    FILE *pf = fopen(sbuf,"r");
    if ( !pf )
    {
        printf("\n\nCould not open %s\n", sbuf);
        return -1;
    }
    int iok, r = 0;
    *sbuf=0;
    do
    {
        iok = fread(sbuf, 1, sizeof(sbuf), pf);
        if (iok>0)
            r+=iok;
    } while (iok<0 && errno==EINTR);
    fclose(pf);
    return r;
#endif
}

double process_mem()
{
    int statlen;
#if defined(__linux__) || defined(__APPLE__)
    static int page_k_shift=0;
    if (page_k_shift==0)
    {
        int j;
        for (j = getpagesize(); j > 1024; j >>= 1)
            page_k_shift++;
    }
#endif
#ifdef __linux__
    statlen = read_file();
    if (statlen>0)
    {
        unsigned long size;
        sscanf(sbuf, "%ld", &size);
        size <<= page_k_shift;
        return ((double)size);
    }
#endif

#ifdef G_PLATFORM_WIN32
    {
        SIZE_T virtualMemUsedByMe;
        PROCESS_MEMORY_COUNTERS_EX pmc;
        GetProcessMemoryInfo(GetCurrentProcess(), (PROCESS_MEMORY_COUNTERS *) &pmc, sizeof(pmc));
        virtualMemUsedByMe = pmc.PrivateUsage;
        //SIZE_T physMemUsedByMe = pmc.WorkingSetSize;
        return ((double) virtualMemUsedByMe);
    }
#endif


#ifdef __APPLE__
    // Inspired by:
    // http://miknight.blogspot.com/2005/11/resident-set-size-in-mac-os-x.html
    gboolean resident=FALSE;
    struct task_basic_info t_info;
    mach_msg_type_number_t t_info_count = TASK_BASIC_INFO_COUNT;
    task_info(current_task(), TASK_BASIC_INFO, (task_info_t)&t_info, &t_info_count);
    double size = 102.4* (resident ? t_info.resident_size : t_info.virtual_size) / getpagesize();
    return size;
#endif

    return 0.0;
}


gboolean
idle_func (GtkWidget *mainW)
{
    static double size_0 = 0.0;
    GObject *O = G_OBJECT(mainW);
    HstHistogram *h1d= HST_HISTOGRAM( g_object_get_data(O,"h1d" ));
    if (h1d)
    {
        double val = randgauss(16000, 8); //25.0 + 0.01 * randuniform();
        val = process_mem() - size_0;
        if (size_0==0.0) {
            size_0 = val;
            val = 0.0;
        }
        hst_histogram_fill(h1d, val/1024/1024, 1);
    }
    
    return TRUE;
}


static gboolean 
delete_event( GtkWidget *widget,
              GdkEvent  *event,
              gpointer   data )
{
    /* Do not destroy with a delete-event */
    return TRUE;
}

int main (int argc, char *argv[]) {
  GtkWidget *mainWindow;
  int iid;

  //std_logger = g_set_printerr_handler( my_logger );
  
  gtk_init (&argc, &argv);
  
  mainWindow = create_mainWindow ();

  //iid = g_idle_add((GSourceFunc)idle_func, mainWindow);
  iid = g_timeout_add(100, (GSourceFunc)idle_func, mainWindow);
  printf ("idle id: %d\n", iid);
  gtk_widget_set_size_request(mainWindow, 300, 250);
  gtk_dialog_run(GTK_DIALOG(mainWindow));
  return 0;
}

GtkWidget *
create_mainWindow (void)
{
    GtkWidget *mainWindow;
    GtkWidget *hst2d, *hst1d;
    HstTracer *tracer;
    
    mainWindow = gtk_dialog_new_with_buttons(
                         "H1D", NULL,
					     GTK_DIALOG_MODAL| GTK_DIALOG_DESTROY_WITH_PARENT,
					     GTK_STOCK_OK,
					     GTK_RESPONSE_ACCEPT,
					     NULL
	);

    tracer = hst_tracer_new(256);
    hst1d = gtk_hst1d_new ( HST_FUNCTION(tracer) );
    gtk_hst1d_set_ytitle(GTK_HST1D(hst1d), "Mb");
    gtk_hst1d_set_xtitle(GTK_HST1D(hst1d), "Sample number");
    g_object_set_data( G_OBJECT(mainWindow),"h1d", tracer);
    
    gtk_container_add (GTK_CONTAINER (  gtk_dialog_get_content_area( GTK_DIALOG(mainWindow) )), hst1d);

    gtk_widget_show_all(mainWindow);

    gtk_hst1d_monitor_start(GTK_HST1D(hst1d),500);

    return mainWindow;
}
